# Copyright 2017 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury.common.clients.router_req_client import RouterReqClient


log = logging.getLogger(__name__)


class RPCFrontEndClient(RouterReqClient):
    """Client for RPC Front End"""

    service_name = 'RPC frontend'

    def get_job(self, job_id, projection=None):
        """

        :param job_id:
        :param projection:
        :return:
        """
        payload = {
            'endpoint': 'get_job',
            'args': [job_id],
            'kwargs': {
                'projection': projection
            }
        }
        return self.transceiver(payload)

    def get_job_status(self, job_id):
        """

        :param job_id:
        :return:
        """
        payload = {
            'endpoint': 'get_job_status',
            'args': [job_id]
        }
        return self.transceiver(payload)

    def get_job_tasks(self, job_id, projection=None):
        """

        :param job_id:
        :param projection:
        :return:
        """
        payload = {
            'endpoint': 'get_job_tasks',
            'args': [job_id],
            'kwargs': {
                'projection': projection
            }
        }

        return self.transceiver(payload)

    def get_task(self, task_id):
        """

        :param task_id:
        :return:
        """
        payload = {
            'endpoint': 'get_task',
            'args': [task_id]
        }
        return self.transceiver(payload)

    def get_jobs(self, projection=None):
        """

        :param projection:
        :return:
        """
        payload = {
            'endpoint': 'get_jobs',
            'kwargs': {
                'projection': projection
            }
        }
        return self.transceiver(payload)

    def create_job(self, query, instruction):
        """

        :param query:
        :param instruction:
        :return:
        """
        payload = {
            'endpoint': 'create_job',
            'args': [query, instruction]
        }
        log.debug('Dispatching query {query} : {instruction}')
        return self.transceiver(payload)



import logging
import time

from multiprocessing.pool import Pool

from mercury.common.clients.router_req_client import RouterReqClient

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

TEST_NO_ASYNC_URL = 'tcp://localhost:9001'
TEST_ASYNC_URL = 'tcp://localhost:9002'

payload = {
    'endpoint': 'enqueue_task',
    'args':
        [{
            'host': 'localhost',
            'port': 9090,
            'method': 'echo',
            'args': ['Hello World!'],
            'kwargs': {},
            'task_id': '1',
            'job_id': '1',
        }]
}


def timer(func):
    def function_timer(*args, **kwargs):
        start = time.time()
        value = func(*args, **kwargs)
        end = time.time()
        runtime = end - start
        log.info(f"{func} took {runtime} seconds to complete")
        return value
    return function_timer


def work(url, n):
    client = RouterReqClient(url)

    results = [client.transceiver(payload) for _ in range(n)]
    client.close()
    return results


@timer
def concurrent_test(c, reqs, use_async):
    pool = Pool(processes=c)

    url = use_async and TEST_ASYNC_URL or TEST_NO_ASYNC_URL

    n = round(reqs / c)

    log.info(f'Starting test [Concurrency: {c}  Requests: {reqs} Async: '
             f'{use_async} ]')
    results = [pool.apply_async(work, (url, n)) for _ in range(c)]

    # wait
    [result.get() for result in results]

    pool.terminate()
    pool.close()


if __name__ == '__main__':
    concurrent_test(10, 1000, use_async=False)
    concurrent_test(10, 1000, use_async=True)

    concurrent_test(15, 2000, use_async=False)
    concurrent_test(15, 2000, use_async=True)



# Copyright 2017 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from gevent.pywsgi import WSGIServer

from mercury.boot.configuration import get_boot_configuration
from mercury.boot.app import app

log = logging.getLogger(__name__)


def main():
    """
    Gevent WSGI server launcher with graceful stop.
    """
    config = get_boot_configuration()
    logging.basicConfig(
        level=logging.getLevelName(config.logging.level),
        format=config.logging.format)
    http_server = WSGIServer((config.host, config.port), app)

    try:
        log.info('Starting gevent WSGI service')
        http_server.serve_forever()
    except KeyboardInterrupt:
        log.info('Stopping gevent WSGI service')
        http_server.stop()


if __name__ == '__main__':
    main()



import re
import json
import time
import operator
from functools import reduce

from behave import given, then, when, step, use_step_matcher
from src.tests.behave.api.features.common.utils import read_json_from_file


@then(
    "the first {service_name} task for the {device_service} device has the stdout output contained in {out_filename}"
)
def step_the_first_task_has_stdout_output_in_file(
    context, service_name, out_filename, device_service
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type out_filename: str
    """
    service_client = context.services[service_name]["client"]
    device_client = context.services[device_service]["client"]
    location = context.json_location
    data = read_json_from_file(out_filename, location)

    tasks = context.services[service_name]["tasks"]
    first_task = tasks[0]

    task_resp = service_client.get(resource_id=first_task["task_id"])
    task_message = task_resp.json()["message"]

    if data["message"] is None:
        context.check.assertEqual(
            data["message"],
            task_message,
            msg="Message was {}, should be {}".format(
                task_message, data["message"]
            ),
        )
    else:
        stdout = data["message"]["stdout"]
        returncode = data["message"]["returncode"]
        stderr = data["message"]["stderr"]

        device_id = context.services[device_service]["id"]
        device_resp = device_client.get(device_id)

        if "_sub_" in stdout:
            # The _sub_ is formated like this _sub_<name_of_data>_sub_
            # The <name_of_data> portion is replaced with the actual value and _sub_ is removed
            # We need to get information from the mercury device details
            # to compare to what was returned from the RPC job

            subs = re.findall("_sub_.*?_sub_", stdout, flags=re.DOTALL)
            for sub in subs:
                keys = sub.replace("_sub_", "").split(".")
                value = reduce(operator.getitem, keys, device_resp.json())
                stdout = stdout.replace(sub, str(value))
            # reduce something like active.rpc_port in order ro read it
            # from a dict, for example resp.json()["active"]["rpc_port"]

        context.check.assertEqual(
            task_message["stdout"],
            stdout,
            msg="Stdout was {}, should be {}".format(
                task_message["stdout"], stdout
            ),
        )
        context.check.assertEqual(
            task_message["returncode"],
            returncode,
            msg="Returncode was {}, should be {}".format(
                task_message["returncode"], returncode
            ),
        )
        context.check.assertEqual(
            task_message["stderr"],
            stderr,
            msg="Stderr was {}, should be {}".format(
                task_message["stderr"], stderr
            ),
        )


@then(
    "the first {service_name} task for the {device_service} device has the message output contained in {out_filename}"
)
def step_the_first_task_has_message_output_in_file(
    context, service_name, out_filename, device_service
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type out_filename: str
    """
    service_client = context.services[service_name]["client"]
    device_client = context.services[device_service]["client"]
    location = context.json_location
    data = read_json_from_file(out_filename, location)

    tasks = context.services[service_name]["tasks"]
    first_task = tasks[0]

    task_resp = service_client.get(resource_id=first_task["task_id"])
    task_message = task_resp.json()["message"]

    if data["message"] is None:
        context.check.assertEqual(
            data["message"],
            task_message,
            msg="Message was {}, should be {}".format(
                task_message, data["message"]
            ),
        )
    else:
        device_id = context.services[device_service]["id"]
        device_resp = device_client.get(device_id)

        task_message_keys = set(task_message.keys())
        device_keys = set(device_resp.json().keys())

        diff = set(
            [item for item in task_message_keys if item not in device_keys]
            + [item for item in device_keys if item not in task_message_keys]
        )
        not_shared = set(
            ["origin", "time_created", "active", "time_updated", "_id"]
        )

        context.check.assertEqual(
            diff, not_shared, msg="Expected keys don't match"
        )

        context.check.assertEqual(
            device_resp.json()["bmc"]["network"]["ip_address"],
            task_message["bmc"]["network"]["ip_address"],
            msg="IP was {}, should be {}".format(
                task_message["bmc"]["network"]["ip_address"],
                device_resp.json()["bmc"]["network"]["ip_address"],
            ),
        )
        # TODO
        # instead of inspector_out.json, have json with list of fields to check
        # similar ot the IP check above



import collections
import copy
import os

from conflagration import api
from conflagration.namespace import ModifiableNamespace
from src.tests.behave.common.const import ConfigVars, Sources


DEFAULT_CONFIG_FILES = (
    #'src/tests/behave/configs/mercury_tests_creds.config',
    #'src/tests/behave/configs/mercury_tests.config',
    # USE ENVIRONMENT VARIABLES
)


def get_conflagration(config_file_paths=DEFAULT_CONFIG_FILES):
    """Instantiate a conflagration instance.

    This defaults to loading from two files:

    1. mercury_tests_creds.config - contains sensitive information
    2. mercury_tests.config - contains non-sensitive information

    This lets us encrypt just the sensitive bits for jenkins and quickly make
    changes to the non-sensitive configs, which we can store in plaintext.
    """
    here = os.getcwd()
    paths = [os.path.join(here, filename) for filename in config_file_paths]
    if config_file_paths:
        cfg = api.conflagration(files=paths, namespace_obj=ModifiableNamespace())
    else:
        cfg = api.conflagration(default_to_env=True)
    return cfg



import time
import json

import requests

from src.tests.behave.api.features.common.exception import TimeoutException


def check_service_api(url, interval_time=5, timeout=20):
    """Check for running service API

    :type url: str
    :type interval_time: int
    :type timeout: int

    """
    end_time = time.time() + timeout

    while time.time() < end_time:
        try:
            resp = requests.get(url)
        except requests.exceptions.RequestException:
            time.sleep(interval_time)
            continue
        if resp.status_code:
            break
    else:
        raise TimeoutException(
            "check_service_api ran for {0} seconds and "
            "could not reach service: {1}".format(timeout, url)
        )
    return resp


def to_singular(name):
    """Convert the name to singular if it is plural

    This just trims a trailing 's', if found.
    """
    return name[:-1] if name.endswith("s") else name


def get_entity_list_container_field(name):
    """Returns the container field used in list responses

    GET /active_computers       -> {"items": [...]}
    GET /jobs                   -> {"jobs": [...]}
    """
    if name == "active_computers":
        return "items"
    elif name == "inventory_computers":
        return "items"
    elif name == "rpc_tasks":
        return "tasks"
    elif name == "rpc_jobs":
        return "jobs"
    return name


def get_entity_id_field(name):
    """Returns the id field used for entities in list responses

    GET /active_computers       -> {"items": [{"mercury_id": "...", ...}}]}
    GET /jobs                   -> {"jobs": [{"job_id": "...", ...}}]}
    """
    if name == "active_computers":
        return "mercury_id"
    elif name == "inventory_computers":
        return "mercury_id"
    elif name == "rpc_tasks":
        return "task_id"
    elif name == "rpc_jobs":
        return "job_id"
    return name


def read_json_from_file(filename, location):
    filename = "{0}/{1}".format(location, filename)
    with open(filename) as file:
        data = json.loads(file.read())
    return data


def wait_for_not_none(func, *args, timeout=50):
    value = func(*args)
    while value == None:
        if timeout <= 0:
            break
        time.sleep(5)
        timeout -= 1
        value = func(*args)
    return value


def keys_in_list_or_dict(key_stack, item):
    if len(key_stack) <= 0:
        return True
    else:
        key = key_stack[0]
        if type(item) == dict:
            if key not in item.keys():
                return False
            new_item = item[key]
            r = keys_in_list_or_dict(key_stack[1:], new_item)
            return r
        elif type(item) == list:
            results = True
            for sub in item:
                r = keys_in_list_or_dict(key_stack, sub)
                results = results and r
            return results
        else:
            return False


def check_params_applied_in_resp(param_data, resp):

    # TODO define checks for specific keys
    # for example the order and limit come back in the respnose

    all_matched = True

    keys = param_data.keys()
    for key in keys:
        # in case a key returned is a different string
        akey = key
        if param_data[key] == None:
            break
        if key == "limit":
            expected_result = param_data[key]
            if type(expected_result) is not int or expected_result < 0:
                expected_result = 250
            actual_result = resp.json()[akey]
            if expected_result != actual_result:
                all_matched = False
        elif key == "sort_direction":
            expected_result = param_data[key]
            if expected_result == -1:
                expected_result = "DESCENDING"
            else:
                expected_result = "ASCENDING"
            actual_result = resp.json()[akey]
            if expected_result != actual_result:
                all_matched = False
        elif key == "projection":
            projections = param_data[key].split(",")
            if "items" in resp.json().keys():
                if resp.json()["items"] == []:
                    return False
                    # TODO add a computer to the DB somehow
                else:
                    for item in resp.json()["items"]:
                        for projection in projections:
                            pstack = projection.split(".")
                            result = keys_in_list_or_dict(pstack, item)
                            all_matched = all_matched and result
            else:
                item = resp.json()
                for projection in projections:
                    pstack = projection.split(".")
                    result = keys_in_list_or_dict(pstack, item)
                    all_matched = all_matched and result

    return all_matched



import asyncio
import logging

import zmq.asyncio

from mercury.common.asyncio.transport import AsyncRouterReqService


log = logging.getLogger('disconnect')

logging.basicConfig(level=logging.DEBUG)


class WorkService(AsyncRouterReqService):
    async def process(self, message):
        if message['action'] == 'sleep':
            await asyncio.sleep(15)
            log.debug('Slept')
            return {'message': 'I slept'}
        if message['action'] == 'fast':
            log.debug('Fast')
            return {'message': 'I was fast'}
        if message['action'] == 'self-destruct':
            for i in reversed(range(6)):
                log.info('Self-destructing in t-minus {} seconds'.format(i))
                await asyncio.sleep(1)
            raise Exception('Have a nice day!')


loop = zmq.asyncio.ZMQEventLoop()
# loop = asyncio.get_event_loop()
loop.set_debug(True)

asyncio.set_event_loop(loop)

service = WorkService('tcp://0.0.0.0:9090', linger=0)

try:
    loop.run_until_complete(service.start())
except KeyboardInterrupt:
    log.info('Shutting down')
    service.kill()
finally:
    pending = asyncio.Task.all_tasks(loop=loop)
    log.debug('Waiting on {} pending tasks'.format(len(pending)))
    loop.run_until_complete(asyncio.gather(*pending))
    log.debug('Shutting down event loop')
    loop.close()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


BACKEND_CONFIG_FILE = 'mercury-backend.yaml'


def add_common_options(configuration):
    configuration.add_option('rpc.db.servers',
                             default='127.0.0.1:27017',
                             special_type=list,
                             help_string='Server or coma separated list of '
                                         'servers to connect to')

    configuration.add_option('rpc.db.name',
                             default='test',
                             help_string='The database for our collections')

    configuration.add_option('rpc.db.replica_name',
                             config_address='rpc.db.replica_name',
                             help_string='Optional replicaset name')

    configuration.add_option('rpc.db.jobs_collection',
                             config_address='rpc.db.jobs_collection',
                             default='rpc_jobs',
                             help_string='The collection for RPC jobs')

    configuration.add_option('rpc.db.tasks_collection',
                             config_address='rpc.db.tasks_collection',
                             default='rpc_tasks',
                             help_string='The collection for RPC tasks')




# Copyright 2018 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from urllib.parse import urlsplit

import pystache


from flask import request, abort, Response
from flask.views import MethodView

from mercury.common.clients.inventory import InventoryClient
from mercury.boot.configuration import get_boot_configuration


log = logging.getLogger(__name__)
configuration = get_boot_configuration()


class BootView(MethodView):
    """
    Boot method view
    """
    def get(self):
        """
        Boot request, transmit the discover script
        :return:
        """

        mercury_boot_url = '://'.join(urlsplit(request.base_url)[:2])

        with open('scripts/discovery.ipxe') as script_file:
            script = pystache.render(script_file.read(), dict(
                mercury_boot_url=mercury_boot_url))

        return Response(script, content_type='text/plain')


class DiscoverView(MethodView):
    """
    Discover method view
    """
    def __init__(self, *args, **kwargs):
        super(DiscoverView, self).__init__(*args, **kwargs)
        inventory_url = configuration.inventory.inventory_router
        self.inventory_client = InventoryClient(inventory_url)

    @staticmethod
    def update_cmdline_backend_options():
        backend_url = configuration.agent.backend_url_cmdline
        log_service_url = configuration.agent.log_service_cmdline

        if backend_url or log_service_url:
            configuration['__BACKEND_OPTIONS__'] = '{}{}'.format(
                ' MERCURY_BACKEND={} '.format(backend_url) if backend_url else ' ',
                'MERCURY_LOG_SERVICE={} '.format(log_service_url) if log_service_url else ''
            )
        else:
            configuration['__BACKEND_OPTIONS__'] = ' '

    @staticmethod
    def render_agent_script():
        with open('scripts/agent.ipxe') as fp:
            template = fp.read()

        DiscoverView.update_cmdline_backend_options()
        return pystache.render(template, **configuration)

    @staticmethod
    def plain(message):
        return Response(message, content_type='text/plain')

    def get(self, mac_address):
        """ Attempt to relate a device using the provided mac address """
        result = self.inventory_client.query({
            'interfaces.address': mac_address
        }, projection={
            'boot': 1,
            'mercury_id': 1,
            'dmi': 1
        })

        if result.get('error'):
            abort(500, result)

        message = result['message']

        if not message['total']:
            log.info('New device: {}'.format(mac_address))
            return self.plain(self.render_agent_script())

        if message['total'] > 1:
            log.error('DUPLICATE MAC ADDRESS: {}'.format(mac_address))
            abort(500, 'Duplicate mac addresses in inventory')

        inventory_data = message['items'][0]

        boot_info = inventory_data.get('boot', {})

        if boot_info.get('script'):
            # Dangerous, can potentially leak entire configuration into image
            # if an attacker knows the key names
            return pystache.render(boot_info['script'], dict(
                **inventory_data, **configuration))

        boot_state = boot_info.get('state', 'agent')

        if boot_state == 'local':
            log.info('Booting {} from hard drive'.format(
                inventory_data['mercury_id']))
            return self.plain('#!ipxe\nexit\n')

        elif boot_info == 'rescue':
            log.info('Booting {} to rescue mode'.format(
                inventory_data['mercury_id']
            ))
            return self.plain('Boot rescue iPXE script here')

        log.info('Booting {} to agent'.format(inventory_data['mercury_id']))
        return self.plain(self.render_agent_script())



import logging
import time

log = logging.getLogger(__name__)


class Worker(object):
    """Worker executing tasks"""
    def __init__(self, task_handler, maximum_requests, maximum_age,
                 handler_args, handler_kwargs):
        """Create a new Worker to execute tasks.

        :param task_handler: Task handler to fetch and execute tasks.
        :param maximum_requests: Maximum number of tasks this worker can
            execute.
        :param maximum_age: Maximum age of the worker.
        """
        self.maximum_requests = maximum_requests
        self.maximum_age = maximum_age
        self.birth = None
        self.kill_signal = None
        self.handled_tasks = 0

        self.task = task_handler(*handler_args, **handler_kwargs)

    def start(self):
        """Start the worker.

        The worker fetches and executes tasks as long as its maximum number
        of tasks is not reached.
        """
        self.birth = time.time()
        while self.handled_tasks < self.maximum_requests:
            if self.kill_signal:
                log.info(
                    'Shutting down... I served my masters well: %s tasks ,'
                    'lived %s seconds' % (self.handled_tasks,
                                          time.time() - self.birth))
                break
            task = self.task.fetch()
            if not task:
                continue
            self.task.execute()
            self.handled_tasks += 1



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
import pymongo
import pytest

from mercury.common import mongo
from tests.common.unit.base import MercuryCommonUnitTest


@mock.patch('pymongo.MongoClient')
def test_get_connection(mock_mongo_client):
    """Test get_connection() creates a new MongoClient."""
    mongo.get_connection('host1')
    mock_mongo_client.assert_called_with(['host1'], replicaset=None)

    mongo.get_connection(['host1', 'host2'], replica_set='replica_set_name')
    mock_mongo_client.assert_called_with(['host1', 'host2'],
                                         replicaset='replica_set_name')


class MongoCollectionUnitTest(MercuryCommonUnitTest):
    def test___init__(self):
        """Test creation of MongoCollection."""
        collection = mongo.MongoCollection('db_name', 'collection_name')

        assert isinstance(collection.db, pymongo.database.Database)
        assert collection.db.name == 'db_name'
        assert isinstance(collection.collection, pymongo.collection.Collection)
        assert collection.collection.full_name == 'db_name.collection_name'

    def test___init__invalid_names(self):
        """Test creation of MongoCollection with invalid names."""
        with pytest.raises(pymongo.errors.InvalidName) as exc:
            collection = mongo.MongoCollection('', 'collection_name')

        with pytest.raises(pymongo.errors.InvalidName):
            collection = mongo.MongoCollection('db_name', '')


def test_get_collection():
    """Test get_collection()"""
    collection = mongo.get_collection('db_name', 'collection_name')

    assert isinstance(collection, pymongo.collection.Collection)
    assert collection.full_name == 'db_name.collection_name'


def test_get_collection_invalid_names():
    """Test get_collection() with invalid names."""
    with pytest.raises(pymongo.errors.InvalidName):
        collection = mongo.get_collection('', 'collection_name')

    with pytest.raises(pymongo.errors.InvalidName):
        collection = mongo.get_collection('db_name', '')



import logging

from mercury.common.asyncio.endpoints import (
    StaticEndpointController, async_endpoint
)

from mercury.common.exceptions import EndpointError, MercuryUserError
from mercury.common.mongo import serialize_object_id

from mercury.rpc.jobs import Job, tasks

log = logging.getLogger(__name__)


class RPCController(StaticEndpointController):
    def __init__(self, inventory_client, jobs_collection, tasks_collection):
        super(RPCController, self).__init__()

        self.inventory_client = inventory_client
        self.jobs_collection = jobs_collection
        self.tasks_collection = tasks_collection

    @staticmethod
    def prepare_for_serialization(obj):
        """Converts object_id to a string and a datetime object to ctime format
        :param obj: probably a task or a job document
        :return: the object reference
        """
        serialize_object_id(obj)
        if obj.get('ttl_time_completed'):
            obj['ttl_time_completed'] = obj['ttl_time_completed'].ctime()
        return obj

    @async_endpoint('get_job')
    async def get_job(self, job_id, projection=None):
        """Gets a job from the job_collection. Jobs expire quickly.

        :param job_id: The Id of the job to get
        :param projection: A mongodb projection. https://goo.gl/kB2g26
        :return: A job object
        """
        job = await self.jobs_collection.find_one({'job_id': job_id},
                                                  projection=projection)
        if not job:
            return

        return self.prepare_for_serialization(job)

    @async_endpoint('get_job_status')
    async def get_job_status(self, job_id):
        """Get the status of job tasks

        :param job_id: the id of a job
        :return: Job object contain task status objects or None
        """
        error_states = ['ERROR', 'TIMEOUT', 'EXCEPTION']
        job = await self.jobs_collection.find_one({'job_id': job_id})
        if not job:
            return

        tasks = self.tasks_collection.find(
            {'job_id': job_id}, {'task_id': 1, 'status': 1, '_id': 0})

        job['has_failures'] = False
        job['tasks'] = []

        async for task in tasks:
            job['tasks'].append(serialize_object_id(task))
            if task['status'] in error_states:
                job['has_failures'] = True

        return self.prepare_for_serialization(job)

    @async_endpoint('get_job_tasks')
    async def get_job_tasks(self, job_id, projection=None):
        """Get tasks belonging to a job

        :param job_id: The id of a job (UUID)
        :param projection: A mongodb projection. https://goo.gl/kB2g26
        :return: dictionary containing top level keys count and jobs
        """
        c = self.tasks_collection.find({'job_id': job_id},
                                       projection=projection)
        count = await c.count()
        tasks = []
        async for task in c:
            tasks.append(self.prepare_for_serialization(task))

        return {'count': count, 'tasks': tasks}

    @async_endpoint('get_task')
    async def get_task(self, task_id):
        """Get a single task

        :param task_id: The id of the task (UUID)
        :return: The task object (dict)
        """

        task = await self.tasks_collection.find_one({'task_id': task_id})

        if not task:
            return

        return self.prepare_for_serialization(task)

    @async_endpoint('get_active_tasks_by_mercury_id')
    async def get_active_tasks_by_mercury_id(self, mercury_id):
        """Gets all active tasks associated with specified mercury id

        :param mercury_id:
        :return: dictionary containing tasks list and count
        """
        c = self.tasks_collection.find({'mercury_id': mercury_id,
                                        'time_completed': None})
        count = await c.count()
        tasks = []
        async for task in c:
            tasks.append(self.prepare_for_serialization(task))

        return {'count': count, 'tasks': tasks}


    @async_endpoint('get_jobs')
    async def get_jobs(self, projection=None):
        """Get active jobs. The jobs collection is made ephemeral via a ttl key;
        this collection should not grow very large

        :param projection: A mongodb projection. https://goo.gl/kB2g26
        :return: dictionary containing top level keys count and jobs
        """
        projection = projection or {'instruction': 0}
        c = self.jobs_collection.find({}, projection=projection).sort(
            'time_created', 1)
        count = await c.count()
        jobs = []
        async for job in c:
            jobs.append(self.prepare_for_serialization(job))
        return {'count': count, 'jobs': jobs}

    @async_endpoint('create_job')
    async def create_job(self, query, instruction):
        """Create a job

        :param query: Query representing targets of the instruction
        :param instruction: An instruction or preprocessor directive. See the
        full documentation regarding instruction syntax at
        http://jr0d.github.io/mercury_api
        :raises EndpointException: Raised after catching a MercuryUserError as
        to conform to dispatch semantics
        :return: The job_id or None
        """

        # Add a constraint to the query that ensures we only match 'active'
        # devices (devices that are accessible through an agent)
        query.update({'active': {'$ne': None}})

        active_targets = await self.inventory_client.query(
            query, projection={'active': 1, 'origin': 1}, limit=0,
            sort_direction=1)

        total_targets = active_targets['message']['total']
        targets = active_targets['message']['items']

        log.debug(f'Found {total_targets} for query {query}')

        if not total_targets:
            return

        try:
            job = Job(instruction, targets, self.jobs_collection,
                      self.tasks_collection)
        except MercuryUserError as mue:
            raise EndpointError(str(mue), 'create_job')

        await job.insert()
        job.enqueue_tasks()

        return {'job_id': str(job.job_id), 'targets': total_targets}

    @async_endpoint('update_task')
    async def update_task(self, update_data):
        self.validate_required(['task_id'], update_data)
        return await tasks.update_task(update_data, self.tasks_collection)

    @async_endpoint('complete_task')
    async def complete_task(self, return_data):
        self.validate_required(['task_id',
                                'job_id',
                                'status',
                                'message'], return_data)
        return await tasks.complete_task(return_data,
                                         self.jobs_collection,
                                         self.tasks_collection)



# Copyright 2018 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury.common.configuration import MercuryConfiguration

BOOT_CONFIG_FILE = 'mercury-boot.yaml'


def options(configuration):
    """ A single place to add program options """

    configuration.add_option(
        'host',
        env_variable='MBOOT_HOST',
        default='127.0.0.1',
        help_string='The host address to bind to')

    configuration.add_option(
        'port',
        env_variable='MBOOT_PORT',
        default=5000,
        special_type=int,
        help_string='The port to bind to')

    configuration.add_option(
        'inventory.inventory_router',
        '--inventory-router',
        default='tcp://127.0.0.1:9000',
        help_string='The inventory router url')

    configuration.add_option(
        'logging.log_file',
        default='mercury-boot.log',
        help_string='The log file path')

    configuration.add_option(
        'logging.level', default='DEBUG', help_string='The app log level')

    configuration.add_option(
        'agent.file_server_url',
        required=True,
        help_string='The file server path containing agent boot files')

    configuration.add_option(
        'agent.kernel',
        default='vmlinuz',
        help_string='The agent kernel')

    configuration.add_option(
        'agent.initrd',
        default='initrd',
        help_string='The agent initial ram file system')

    configuration.add_option(
        'agent.rootfs_option',
        help_string='The kernel command line option used by the initrd to find '
                    'the root file system',
        required=True
    )

    configuration.add_option(
        'agent.file_system',
        default='mercury-agent.sqfs',
        help_string='The agent file system'
    )

    configuration.add_option(
        'agent.kernel_options',
        default='',
        help_string='Kernel options for the agent environment'
    )

    configuration.add_option(
        'agent.backend_url_cmdline',
        help_string='Provide an alternative backend url. This will be appended to the agent '
                    'kernel command line, overriding the configuration baked into the image'
    )

    configuration.add_option(
        'agent.log_service_cmdline',
        help_string='Provide an alternative log service url. This will be appended to the agent '
                    'kernel command line, overriding the configuration baked into the image'
    )


def get_boot_configuration():
    boot_configuration = MercuryConfiguration('mercury-boot', BOOT_CONFIG_FILE)
    options(boot_configuration)

    return boot_configuration.scan_options()



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from mercury.common.clients import inventory
from mercury.common.exceptions import MercuryClientException

from tests.common.unit import base


class FakeInventoryClient(inventory.InventoryClient):
    service_name = 'Inventory'
    """Helper class to fake an inventory client in tests."""
    def __init__(self):
        self.ctx = None
        self.socket = None


class InventoryClientUnitTest(base.MercuryCommonUnitTest):
    """Tests for mercury.common.inventory_client.client.InventoryClient"""
    def setUp(self):
        super(InventoryClientUnitTest, self).setUp()
        self.inventory_client = FakeInventoryClient()

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_insert_one(self, mock_transceiver):
        """Test insert_one()."""
        mock_transceiver.return_value = {'response': 'fake_response'}

        device_info = {'mercury_id': 1}
        payload = {
            'endpoint': 'insert_one',
            'args': [device_info]
        }

        response = self.inventory_client.insert_one(device_info)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])

    def test_insert_one_no_mercury_id(self):
        """Test insert_one() fails when no mercury_id in device_info."""
        device_info = {}
        self.assertRaises(MercuryClientException,
                          self.inventory_client.insert_one,
                          device_info)

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_update_one(self, mock_transceiver):
        """Test update_one()."""
        mock_transceiver.return_value = {'response': 'fake_response'}
        mercury_id = 1
        update_data = 'fake_data'
        payload = {
            'endpoint': 'update_one',
            'args': [mercury_id],
            'kwargs': {'update_data': update_data}
        }

        response = self.inventory_client.update_one(mercury_id, update_data)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_get_one(self, mock_transceiver):
        """Test get_one()."""
        mock_transceiver.return_value = {'response': 'fake_response'}
        mercury_id = 1
        payload = {
            'endpoint': 'get_one',
            'args': [mercury_id],
            'kwargs': {'projection': None}
        }

        response = self.inventory_client.get_one(mercury_id)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_query(self, mock_transceiver):
        """Test query()."""
        mock_transceiver.return_value = {'response': 'fake_response'}
        query_data = 'fake_data'
        payload = {
            'endpoint': 'query',
            'args': [query_data],
            'kwargs': {
                'projection': None,
                'limit': 0,
                'sort': '_id',
                'sort_direction': 1,
                'offset_id': None
            }
        }

        response = self.inventory_client.query(query_data)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_delete(self, mock_transceiver):
        """Test delete()."""
        mock_transceiver.return_value = {'response': 'fake_response'}
        mercury_id = 1
        payload = {
            'endpoint': 'delete',
            'args': [mercury_id],
        }

        response = self.inventory_client.delete(mercury_id)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])

    @mock.patch('mercury.common.clients.router_req_client.RouterReqClient.transceiver')
    def test_count(self, mock_transceiver):
        """Test count()."""
        mock_transceiver.return_value = {'response': 'fake_response'}
        query_data = 'fake_data'
        payload = {
            'endpoint': 'count',
            'args': [query_data],
        }

        response = self.inventory_client.count(query_data)
        mock_transceiver.assert_called_once_with(payload)
        self.assertEqual('fake_response', response['response'])



import json
import random
from behave import when, step
from src.tests.behave.api.features.common.utils import (
    get_entity_list_container_field,
    get_entity_id_field,
    read_json_from_file,
)


@step("a {service_name} entity id is located for testing")
def step_a_service_id_is_located_for_testing(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    # Find a valid service entity from list call
    context.services[service_name]["id"] = None
    service_client = context.services[service_name]["client"]
    response = service_client.get()

    container_field = get_entity_list_container_field(service_name)
    field_name = get_entity_id_field(service_name)
    try:
        # TODO config value?
        service_entities = response.json()[container_field]
        # TODO if this is empty add one somehow?
        context.check.assertGreater(len(service_entities), 0)
        i = random.randint(0, len(service_entities) - 1)
        entity_id = service_entities[i][field_name]
        context.services[service_name]["id"] = entity_id
    except IndexError:
        context.check.assertIsNotNone(
            context.services[service_name]["id"],
            msg="WIP: Create {} API Not Yet Implemented".format(service_name),
        )
        # TODO: Create new service entity since we don't have any


@step(
    "first device queried from {filename} of type {service_name} entity id is located for testing"
)
def step_a_service_id_is_located_for_testing(context, service_name, filename):
    """
    :type context: behave.runner.Context
    :type filename: str
    :type service_name: str
    """

    location = context.json_location
    data = read_json_from_file(filename, location)

    service_client = context.services[service_name]["client"]

    response = service_client.post(data=json.dumps(data), url_suffix="query")

    container_field = get_entity_list_container_field(service_name)
    field_name = get_entity_id_field(service_name)
    try:
        # TODO config value?
        service_entities = response.json()[container_field]
        # TODO if this is empty add one somehow?
        context.check.assertGreater(len(service_entities), 0)
        entity_id = service_entities[0][field_name]
        context.services[service_name]["id"] = entity_id
    except IndexError:
        context.check.assertIsNotNone(
            context.services[service_name]["id"],
            msg="WIP: Create {} API Not Yet Implemented".format(service_name),
        )
        # TODO: Create new service entity since we don't have any


@step("an active {service_name} entity id is located for testing")
def step_a_service_id_is_located_for_testing(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    # Find a valid service entity from list call
    context.services[service_name]["id"] = None
    service_client = context.services[service_name]["client"]
    response = service_client.get()

    container_field = get_entity_list_container_field(service_name)
    field_name = get_entity_id_field(service_name)
    try:
        # TODO config value?
        service_entities = response.json()[container_field]
        # TODO if this is empty add one somehow?
        context.check.assertGreater(len(service_entities), 0)
        i = random.randint(0, len(service_entities) - 1)
        entity_id = service_entities[i][field_name]
        resp = service_client.get(entity_id)

        while resp.json()["active"] == None:
            i += 1
            entity_id = service_entities[i][field_name]
            resp = service_client.get(entity_id)

        context.services[service_name]["id"] = entity_id
    except IndexError:
        context.check.assertIsNotNone(
            context.services[service_name]["id"],
            msg="WIP: Create {} API Not Yet Implemented".format(service_name),
        )
        # TODO: Create new service entity since we don't have any


@step("a {service_name} entity id is defined for testing")
def step_a_test_id_is_defined_for_testing(context, service_name):
    """
    :type context: behave.runner.Context
    """
    context.services[service_name]["id"] = "1234"


@when("I get the entity using the {service_name} api")
def step_i_get_the_entity_using_the_service_api(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(
        context.services[service_name]["id"]
    )


@when(
    "I get with bad headers in {filename} the entity using the {service_name} api"
)
def step_i_get_the_entity_using_the_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """

    location = context.json_location
    headers = read_json_from_file(filename, location)

    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(
        context.services[service_name]["id"], headers=headers
    )


@when(
    "I get with parameters in {filename} the entity using the {service_name} api"
)
def step_i_get_the_entity_with_params_in_filename(
    context, service_name, filename
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """
    location = context.json_location
    data = read_json_from_file(filename, location)

    keys = data.keys()
    suffix = "?"
    for key in keys:
        suffix = "{0}{1}={2}&".format(suffix, key, data[key])
    # trim trailing &
    suffix = suffix.rstrip("&")
    context.services[service_name]["param_data"] = data
    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(
        resource_id=context.services[service_name]["id"], url_suffix=suffix
    )


@when("I get the status of the entity using the {service_name} api")
def step_i_get_the_status_of_the_entity_using_the_service_api(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(
        context.services[service_name]["id"], url_suffix="status"
    )


@when(
    "I get the {tasks_service_name} tasks of the entity using the {service_name} api"
)
def step_i_get_the_tasks_of_the_entity_using_the_service_api(
    context, tasks_service_name, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    context.services[tasks_service_name]["resp"] = service_client.get(
        context.services[service_name]["id"], url_suffix="tasks"
    )


@when(
    "I get a task from the {service_name} entity using the {tasks_service_name} api"
)
def step_i_get_a_task_from_the_entity_using_the_service_api(
    context, service_name, tasks_service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    tasks_service_client = context.services[tasks_service_name]["client"]
    tasks_resp = context.services[tasks_service_name]["resp"]

    first_task = tasks_resp.json()["tasks"][0]
    # TODO config value?
    task_id = first_task["task_id"]
    context.services[tasks_service_name]["id"] = task_id
    context.services[tasks_service_name]["resp"] = tasks_service_client.get(
        resource_id=task_id
    )


@when(
    "I get with bad headers in {filename} a task from the {service_name} entity using the {tasks_service_name} api"
)
def step_i_get_a_task_from_the_entity_using_the_service_api(
    context, service_name, tasks_service_name, filename
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type tasks_service_name: str
    :type filename: str
    """
    location = context.json_location
    headers = read_json_from_file(filename, location)

    service_client = context.services[service_name]["client"]
    tasks_service_client = context.services[tasks_service_name]["client"]
    tasks_resp = context.services[tasks_service_name]["resp"]

    first_task = tasks_resp.json()["tasks"][0]
    # TODO config value?
    task_id = first_task["task_id"]
    context.services[tasks_service_name]["id"] = task_id
    context.services[tasks_service_name]["resp"] = tasks_service_client.get(
        resource_id=task_id, headers=headers
    )


@when(
    "I get a task by id from the {service_name} entity using the {tasks_service_name} api"
)
def step_i_get_a_task_by_id_from_the_entity_using_the_service_api(
    context, service_name, tasks_service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    tasks_service_client = context.services[tasks_service_name]["client"]

    task_id = context.services[tasks_service_name]["id"]
    context.services[tasks_service_name]["resp"] = tasks_service_client.get(
        resource_id=task_id
    )


@step("the {service_name} response contains valid single entity details")
def step_the_service_response_contains_valid_single_entity_details(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_resp = context.services[service_name]["resp"]
    service_entity = service_resp.json()
    context.check.assertIsInstance(service_entity, dict)
    expected_id = context.services[service_name]["id"]
    actual_id = service_entity.get(get_entity_id_field(service_name))
    context.check.assertEqual(actual_id, expected_id)
    # TODO: validate actual content of entity
    # TODO: need to be able to tell what the entity is, in some
    # tests it's a job, some it's a task, some it's a computer


@step("the {service_name} response contains valid single entity status details")
def step_the_service_response_contains_valid_single_entity_status_details(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_resp = context.services[service_name]["resp"]
    service_entity = service_resp.json()
    context.check.assertIsInstance(service_entity, dict)
    # TODO: validate actual content of entity


@step("I have a {invalid_id} for {tasks_service_name}")
def step_have_invalid_id(context, invalid_id, tasks_service_name):
    """
    :type context: behave.runner.context
    :type invalid_id: str
    :type tasks_service_name: str
    """
    context.services[tasks_service_name]["id"] = invalid_id



import logging
import time

from mercury.common.configuration import MercuryConfiguration
from mercury.common.exceptions import MercuryClientException
from mercury.common.mongo import get_collection, get_connection
from mercury.common.transport import SimpleRouterReqService

LOG = logging.getLogger(__name__)

MERCURY_LOG_CONFIG = 'mercury-log.yaml'


def options():
    configuration = MercuryConfiguration(
        'mercury-log',
        MERCURY_LOG_CONFIG,
        description='The mercury logging backend')

    configuration.add_option('log_service.bind_address',
                             default='tcp://127.0.0.1:9006',
                             help_string='The address to bind to'
                             )

    configuration.add_option('log_service.db.servers',
                             default='127.0.0.1:27017',
                             special_type=list,
                             help_string='Server or coma separated list of '
                                         'servers to connect to')

    configuration.add_option('log_service.db.name',
                             config_address='log_service.db.name',
                             default='test',
                             help_string='The database for our collections')

    configuration.add_option('log_service.db.collection',
                             default='log',
                             help_string='The collection for our documents')

    configuration.add_option('log_service.db.replica_name',
                             help_string='An optional replica')

    return configuration.scan_options()


class AgentLogService(SimpleRouterReqService):
    """
    Logging aggregation end point for MercuryAgents
    """

    def __init__(self, bind_address, log_collection):
        super(AgentLogService, self).__init__(bind_address)
        self.log_collection = log_collection

    @staticmethod
    def validate_message(message):
        LOG.debug(message)
        required = [
            'levelno',
            'pathname',
            'message',
            'name'
        ]

        for req in required:
            if req not in message:
                return False

        return True

    @staticmethod
    def set_job_info_from_thread(message):
        """
        The task runner thread (agent.task_runner) has the following naming
        convention:

        _<job_id>_<task_id>

        This lets us associate logging messages to jobs/tasks from within the
        execution thread.
        :param message: reference to the log message
        :return: None
        """
        thread_name = message.get('threadName')
        if thread_name and thread_name[0] == '_':
            job_id, task_id = thread_name.split('_')[1:]
            message['job_id'] = job_id
            message['task_id'] = task_id

    def process(self, message):
        if not self.validate_message(message):
            raise MercuryClientException('Invalid message')

        message.update({'time_created': time.time()})

        self.set_job_info_from_thread(message)

        self.log_collection.insert(message)

        return {'message': 'ok', 'error': False}


def main():
    config = options()
    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)

    db_connection = get_connection(config.log_service.db.servers,
                                   config.log_service.db.replica_name)

    collection = get_collection(config.log_service.db.name,
                                config.log_service.db.collection,
                                db_connection)

    agent_log_service = AgentLogService(config.log_service.bind_address,
                                        collection)

    LOG.info('Starting logging backend on {}'.format(
        config.log_service.bind_address))

    agent_log_service.start()


if __name__ == '__main__':
    main()



import json
import time
import operator
from functools import reduce

from behave import given, when, step
from src.tests.behave.api.features.common.utils import (
    get_entity_list_container_field,
    get_entity_id_field,
    read_json_from_file,
    wait_for_not_none,
)


@given(
    "I have job injection details in {filename} for creating jobs using the {service_name} api"
)
def step_i_have_job_injection_details_in_filename_for_creating_jobs_using_the_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    location = context.json_location
    data = read_json_from_file(filename, location)

    context.services[service_name]["details"]["job_data"] = data


@given(
    "I have job injection details for a specific {device_service} device in {filename} for creating jobs using the {service_name} api"
)
def step_i_have_job_injection_details_for_specific_device_in_filename_for_creating_jobs_using_the_service_api(
    context, filename, service_name, device_service
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    location = context.json_location
    data = read_json_from_file(filename, location)
    data["query"]["mercury_id"] = context.services[device_service]["id"]

    context.services[service_name]["details"]["job_data"] = data


@when("I get the injection results from a post to {service_name}")
def step_i_get_the_injection_results_from_a_post_to_service_api(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["job_data"]

    context.services[service_name]["resp"] = service_client.post(
        data=json.dumps(data)
    )


@when(
    "I get with bad headers in {filename} the injection results from a post to {service_name}"
)
def step_i_get_with_bad_headers_the_results_from_a_post_to_service(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """

    location = context.json_location
    headers = read_json_from_file(filename, location)

    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["job_data"]

    context.services[service_name]["resp"] = service_client.post(
        data=json.dumps(data), headers=headers
    )


@step("the response contains a {service_name} job_id")
def step_the_response_contains_a_job_id(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_resp = context.services[service_name]["resp"]
    job_id = service_resp.json()["job_id"]
    context.check.assertIsInstance(job_id, str)
    context.services[service_name]["id"] = job_id
    # TODO is the job id valid?


@step(
    "the corresponding {service_name} job is completed with {state} state {tasks_service_name} tasks"
)
def step_the_corresponding_service_job_is_completed_and_successful(
    context, service_name, tasks_service_name, state
):
    """
    :type context: behave.runner.Context
    """
    service_client = context.services[service_name]["client"]
    service_resp = service_client.get(context.services[service_name]["id"])
    context.services[service_name]["resp"] = service_resp

    def get_time_completed(job_id):
        return service_client.get(job_id).json()["ttl_time_completed"]

    completed = wait_for_not_none(
        get_time_completed, context.services[service_name]["id"]
    )

    status_resp = service_client.get(
        context.services[service_name]["id"], url_suffix="status"
    )
    tasks = status_resp.json()["tasks"]
    context.services[tasks_service_name]["tasks"] = tasks
    statuses = [task["status"] for task in tasks]
    for status in statuses:
        context.check.assertEqual(status, state)
    has_failures = status_resp.json()["has_failures"]
    context.check.assertEqual(has_failures, False)

    # TODO is the job valid/working?


# Used for bad method testing
@when("I inject with {method} for {service_name}")
def step_i_use_method_on_inject_service(context, method, service_name):
    """
    :type context: behave.runner.Context
    :type method: str
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["job_data"]

    if method == "delete":
        context.services[service_name]["resp"] = service_client.delete()
    # TODO etc



# Copyright 2017 Ruben Quinones (ruben.quinones@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging


log = logging.getLogger(__name__)


class HookException(Exception):
    pass


class Hook(object):
    """
    Base Hook class. Hooks provide a way to manage inventory data before or
    after it is inserted or updated in the database and run additional 
    processes if needed.
    """

    def __init__(self, data, *args, **kwargs):
        self.data = data

    def run(self):
        pass


class InterfaceHook(Hook):
    """
    Interface data handler, this should run when updating interface data for
    a given inventory record.
    """

    def process_data(self):
        """
        Convert interface data to mongodb dot notation in order to update the 
        inventory record without erasing data collected asynchronously.
        """
        interfaces = self.data.pop('interfaces', [])
        for i, interface in enumerate(interfaces):
            for key, value in interface.items():
                new_key = 'interfaces.{}.{}'.format(i, key)
                self.data[new_key] = value

    def run(self):
        self.process_data()


class LLDPHook(Hook):
    """
    Interface LLDP data handler, this should run when updating LLDP data for
    a given inventory interface.
    """

    def process_data(self):
        """
        Convert interface LLDP data to mongodb dot notation in order to update
        the interface specified by the 'interface_index' key.
        """
        try:
            lldp = self.data.pop('lldp')
            interface_index = lldp.pop('interface_index')
        except KeyError:
            raise HookException('Missing LLDP data')
        lldp_key = 'interfaces.{}.lldp'.format(interface_index)
        self.data[lldp_key] = lldp

    def run(self):
        self.process_data()


HOOK_MAP = {
    'interfaces': InterfaceHook,
    'lldp': LLDPHook,
}


def run_hooks(hooks, data):
    """
    Calls the run method for each hook in the hooks dict.
    
    :param hooks: A dict of keys/hook classes
    :param data: Inventory data dict
    :return: 
    """
    for hook_key, hook_class in hooks.items():
        hook = hook_class(data)
        log.info('Running {} hook'.format(hook_key))
        hook.run()


def get_hooks_from_data(data):
    """
    Returns a dict of hooks based on the keys present in the data dict
    
    :param data: Inventory data dict
    :return: dict 
    """
    keys = set(data.keys()) & set(HOOK_MAP.keys())
    hooks = {key: HOOK_MAP[key] for key in keys}
    return hooks



from collections import defaultdict

from behave import given, then, when, step, use_step_matcher

from src.tests.behave.api.features.common.utils import (
    check_params_applied_in_resp,
)
from src.tests.behave.api.features.client import APIClient


use_step_matcher("parse")


@given("the account is an {auth} tenant")
def step_the_account_is_an_auth_tenant(context, auth):
    """
    :type context: behave.runner.Context
    :type auth: str
    """
    context.authorized = True
    if auth == "unauthorized":
        context.authorized = False


@given("the auth token for the {service_name} client is nonexistent")
def step_the_auth_token_is_nonexistent(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    client = context.services[service_name]["client"]
    client.headers.pop("X-Auth-Token")


@given("a {service_name} bad url {bad_url} is provided")
def step_a_service_bad_url_is_provided(context, service_name, bad_url):
    """
    : type context: behave.runner.Context
    : type service_name: str
    : type bad_url: str
    """
    request_kwargs = dict()
    request_kwargs["timeout"] = None
    request_kwargs["ssl_certificate_verify"] = False
    request_kwargs["verbose"] = True

    client_kwargs = dict()
    client_kwargs["authorized"] = context.authorized

    context.services[service_name]["name"] = service_name

    full_service_url = context.base_url + bad_url
    context.services[service_name]["url"] = full_service_url
    context.services[service_name]["client"] = APIClient(
        base_url=full_service_url,
        request_kwargs=request_kwargs,
        client_kwargs=client_kwargs,
    )


@given("the {service_name} client URL is {service_url}")
def step_the_service_client_url_is(context, service_name, service_url):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type service_url: str
    """
    request_kwargs = dict()
    request_kwargs["timeout"] = None
    request_kwargs["ssl_certificate_verify"] = False
    request_kwargs["verbose"] = True

    client_kwargs = dict()
    client_kwargs["authorized"] = context.authorized

    # Keep track of services being used
    context.services[service_name] = defaultdict(dict)
    context.services[service_name]["name"] = service_name

    # Construct complex urls
    if "_id>" in service_url:
        new_url_parts = service_url.split("/")
        url_parts = service_url.split("/")
        for index, element in enumerate(url_parts):
            # note:  all feature files will have to conform to using
            # *_id> in any url passed in for service entity ids
            # ex: /loadbalancers/<lb_id>/nodes/<node_id>
            if "_id>" in element:
                service_url_part = url_parts[index - 1]
                new_url_parts[index] = context.services[service_url_part]["id"]
        service_url = "/".join(new_url_parts)

    full_service_url = context.base_url + service_url
    context.services[service_name]["url"] = full_service_url
    context.services[service_name]["client"] = APIClient(
        base_url=full_service_url,
        request_kwargs=request_kwargs,
        client_kwargs=client_kwargs,
    )


@then("the {service_name} response status is {status_code} {reason}")
def step_the_service_response_status_is(
    context, service_name, status_code, reason
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type status_code: str
    :type reason: str
    """
    context.check.assertEqual(
        context.services[service_name]["resp"].status_code,
        int(status_code),
        msg="Response status code was {}, should be {}".format(
            context.services[service_name]["resp"], status_code
        ),
    )

    actual_resp_reason = context.services[service_name]["resp"].reason
    context.check.assertEqual(
        actual_resp_reason.upper(),
        reason.upper(),
        msg="Response reason was {}, should be {}".format(
            actual_resp_reason, reason
        ),
    )


@step("the {service_name} response contains an error message of")
def step_the_service_response_contains_an_error_message_of(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    # replace line breaks with double space in expected message
    expected_error_message = context.text.replace("\r", "  ").replace(
        "\n", "  "
    )
    actual_error_message = context.services[service_name]["resp"].json()[
        "message"
    ]
    context.check.assertEqual(
        actual_error_message,
        expected_error_message,
        msg="Response message was {}, should be {}".format(
            actual_error_message, expected_error_message
        ),
    )


@step("a {service_name} {invalid_id} is provided")
def step_a_service_invalid_id_is_provided(context, service_name, invalid_id):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type invalid_id: str
    """
    context.services[service_name]["id"] = invalid_id


@then("url parameters to the {service_name} api are applied")
def step_url_parameters_to_service_api_are_applied(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """

    resp = context.services[service_name]["resp"]
    param_data = context.services[service_name]["param_data"]
    params_applied = check_params_applied_in_resp(param_data, resp)

    context.check.assertTrue(params_applied)


@then("the valid url parameters to the {service_name} api are applied")
def step_valid_url_parameters_to_service_api_are_applied(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """

    resp = context.services[service_name]["resp"]
    param_data = context.services[service_name]["param_data"]
    keys = param_data.keys()

    # Check each url parameter
    # If null was passed as a url param, ignore it
    # Remove any invalid projection keys so we don't check for step_i_get_the_list_of_service
    # Since they are ignored when the query is passed to mongo
    for key in keys:
        if param_data[key] == None:
            break
        if key == "projection":
            p = param_data[key].split(",")
            p = [projection for projection in p if "invalid" not in projection]
            for projection in p:
                if "invalid" in projection:
                    p.remove(projection)
            param_data[key] = ",".join(p)
    # Check that all the (valid) url parameters are applied properly
    params_applied = check_params_applied_in_resp(param_data, resp)
    context.check.assertTrue(params_applied)


# Used for bad method testing
@when("I use {method} on {service_name}")
def step_i_use_method_on_service(context, method, service_name):
    """
    :type context: behave.runner.Context
    :type method: str
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    if method == "get":
        context.services[service_name]["resp"] = service_client.get()
    elif method == "post":
        context.services[service_name]["resp"] = service_client.post(data={})
    # TODO etc


# Used for bad method testing
@when("I with an entity use {method} on {service_name}")
def step_i_use_method_on_service_with_entity(context, method, service_name):
    """
    :type context: behave.runner.Context
    :type method: str
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    if method == "get":
        context.services[service_name]["resp"] = service_client.get(
            context.services[service_name]["id"]
        )
    elif method == "post":
        context.services[service_name]["resp"] = service_client.post(
            data={}, resource_id=context.services[service_name]["id"]
        )



"""
environment.py

This file is required by Behave to be in the same directory as where the
feature file is being run.

Because we want to be able to reuse the settings in an environment file,
it is also defined in the parent directory.

Each new service will have a it's own environment file which would wildcard
import the methods from the parent file.

see: https://media.readthedocs.org/pdf/behave/latest/behave.pdf
"""
from src.tests.behave.api.features.environment import (  # noqa # pylint: disable=unused-import
    before_all,
    after_all,
    before_feature,
    after_feature,
    before_scenario,
    after_scenario,
    before_step,
    after_step,
)



import logging
import time

from mercury.common.exceptions import MercuryCritical, MercuryGeneralException

from mercury.inventory.hooks import get_hooks_from_data, run_hooks

log = logging.getLogger(__name__)


class InventoryDBController(object):
    def __init__(self, collection):
        self.collection = collection
        self.collection.create_index('mercury_id', unique=True)

    async def insert_one(self, data):
        mercury_id = data.get('mercury_id')
        if not mercury_id:
            raise MercuryCritical(
                'MercuryID is missing from payload. Shame. Shame. Shame.')

        existing = await self.collection.find_one({'mercury_id': mercury_id},
                                                  projection={'_id': 1})
        if existing:
            # implicit update
            log.info('Record exists, performing update instead.')
            del data['mercury_id']
            await self.update_one(mercury_id, data)
            return existing['_id']

        now = time.time()
        data['time_created'] = now
        data['time_updated'] = now

        insert_result = await self.collection.insert_one(data)
        return insert_result.inserted_id

    async def update_one(self, mercury_id, data=None):
        """
        Very simple update_one/insert method. Here monitor hooks will be run.

        Monitor hooks will inspect the update data for information they care about and perform
        some sort of action. For instance, one such hook may monitor for an update containing
        LLDP information. The hook might publish this transaction to a subscriber interested in
        such things.

        :param mercury_id: The update target
        :param data: The update or insert data
        :return:
        """
        data = data or {}
        upsert = False

        if not data:
            log.warning('Update data is empty')

        if not await self.collection.count({'mercury_id': mercury_id},
                                           projection={'_id': 1}):
            raise MercuryGeneralException(
                'Attempting to update non-existing record')

        if 'mercury_id' in data:
            raise MercuryCritical(
                'MercuryID cannot be updated once a record already exists')

        hooks = get_hooks_from_data(data)
        if hooks:
            # We care about how this data is handled
            upsert = True
            run_hooks(hooks, data)

        data['time_updated'] = time.time()
        await self.collection.update_one({'mercury_id': mercury_id},
                                         {'$set': data}, upsert=upsert)

    async def delete(self, mercury_id):
        log.info('Deleting: %s' % mercury_id)
        await self.collection.delete_one({'mercury_id': mercury_id})

    def get_one(self, mercury_id, projection=None):
        log.debug('Fetching: %s' % mercury_id)
        return self.collection.find_one({'mercury_id': mercury_id},
                                        projection=projection)

    # TODO: Ditch query and add find() and find_one()
    def query(self, query, extra_projection=None, limit=0, sort='_id',
              sort_direction=1, offset_id=None):
        log.debug(
            'Executing query: {} limit: {} key: {} direction: {}'.format(
                query,
                limit,
                sort,
                sort_direction))

        projection = {'mercury_id': 1}

        if extra_projection:
            projection.update(extra_projection)

        if offset_id and sort not in query:
            inequality_operator = '$gt' if sort_direction >= 0 else '$lt'
            query.update({sort: {inequality_operator: offset_id}})

        c = self.collection.find(query, projection=projection)

        if limit:
            c.limit(limit)

        if sort != '_id' or sort_direction != 1:
            # non-default sort
            c.sort(sort, sort_direction)

        return c

    def count(self, query):
        return self.collection.count(query)



import unittest

from mock import patch

from mercury.inventory import hooks


class MercuryInventoryHookTests(unittest.TestCase):
    """
    General hook unit tests.
    """

    def test_get_hooks_from_data(self):
        data = {'lldp': None, 'random': 1234}
        hook_dict = hooks.get_hooks_from_data(data)
        self.assertDictEqual(hook_dict, {'lldp': hooks.LLDPHook})

    @patch('mercury.inventory.hooks.LLDPHook.run')
    def test_run_hooks(self, lldp_run):
        lldp_run.return_value = None
        data = {'lldp': None, 'random': 1234}
        hook_dict = hooks.get_hooks_from_data(data)
        hooks.run_hooks(hook_dict, data)
        lldp_run.assert_called_once()


class MercuryInventoryLLDPHookTests(unittest.TestCase):
    """
    LLDP hook unit tests.
    """

    def test_lldp_hook(self):
        data = {
            'lldp': {
                'switch_name': 'someswitch',
                'switch_port': 'someport',
                'interface_index': 0,
            }
        }

        lldp_hook = hooks.LLDPHook(data)
        lldp_hook.run()

        expected_data = {
            'interfaces.0.lldp': {
                'switch_name': 'someswitch',
                'switch_port': 'someport',
            }
        }

        self.assertDictEqual(data, expected_data)

    def test_lldp_hook_exception(self):
        data = {
            'lldp': {
                'switch_name': 'someswitch',
                'switch_port': 'someport',
            }
        }

        lldp_hook = hooks.LLDPHook(data)

        self.assertRaises(hooks.HookException, lldp_hook.run)


class MercuryInventoryInterfaceHookTests(unittest.TestCase):
    """
    Interface hook unit tests.
    """

    def test_interface_hook(self):
        data = {
            'interfaces': [
                {
                    'name': 'eno1',
                    'subdict': {'subkey': None},
                }
            ]
        }

        interface_hook = hooks.InterfaceHook(data)
        interface_hook.run()

        expected_data = {
            'interfaces.0.name': 'eno1',
            'interfaces.0.subdict': {'subkey': None},
        }

        self.assertDictEqual(data, expected_data)

    def test_interface_hook_is_empty(self):
        data = {}
        interface_hook = hooks.InterfaceHook(data)
        interface_hook.run()
        self.assertEqual(data, {})



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pystache
import yaml

from mercury.common.exceptions import MercuryUserError
from mercury.rpc.preprocessors import preprocessor


def __asset_list_hax(assets):
    """Converts any lists to a string so they can be properly rendered by pystache
    """
    for k, v in list(assets.items()):

        if isinstance(v, dict):
            __asset_list_hax(v)

        if isinstance(v, list):
            assets[k] = '[%s]' % ', '.join(v)


@preprocessor.preprocessor('press_static_assets',
                           'Uses user supplied assets to render press configuration templates')
def press_static_assets(target, instruction):
    """Uses a mercury_id indexed asset store which is supplied, in it's entirety, within the instruction
    :param target: A target containing a mercury_id
    :param instruction: A dictionary containing two fields, template and assets
        template is a string containing a yaml formatted press configuration mustache template
        assets is a dictionary indexed by mercury_id. Each value contains render information
        relevant to the template. If the asset data does not contain data for the target mercury_id
        a MercuryUserException exception is raised

        Note, this is throw away code. The mercury_assets backend will take over the functionality provided
        herein.
    :return: exec_press capability contract
    """

    template = '\n'.join(instruction.get('template'))
    asset_db = instruction.get('assets')

    if not template and asset_db:
        raise MercuryUserError('Contract is incomplete')

    render_data = asset_db.get(target['mercury_id'])

    if not render_data:
        raise MercuryUserError('Assets supplied do not cover target')

    __asset_list_hax(render_data)

    rendered = pystache.render(template, **render_data)

    configuration = yaml.load(rendered)

    return {'method': 'press', 'kwargs': {'configuration': configuration}}



import asyncio
import logging
import time

import msgpack
import zmq
import zmq.asyncio

from mercury.common.asyncio.clients.inventory import InventoryClient

log = logging.getLogger(__name__)

active_state = {}
stop_ping_loop = False


def stop_ping():
    global stop_ping_loop
    stop_ping_loop = True


def add_active_record(record):
    log.info('Adding record, {mercury_id}, to active state'.format(**record))
    active_state.update({
        record['mercury_id']: {
            'mercury_id': record['mercury_id'],
            'rpc_address': record['active']['rpc_address'],
            'rpc_address6': record['active']['rpc_address6'],
            'ping_port': record['active']['ping_port'],
            'last_ping': 0,
            'pinging': False
        }
    })


async def ping(record, ctx, timeout, retries, backoff, inventory_client, rpc_client):
    """
    ping node until it responds or encounters x timeouts of x*backoff
    :param timeout:
    :param backoff:
    :param retries:
    :param record:
    :param ctx:
    :param inventory_client:
    :return:
    """
    failures = 0

    zurl = 'tcp://{}:{}'.format(record['rpc_address6'] or record['rpc_address'],
                                record['ping_port'])

    while failures < retries:
        socket = ctx.socket(zmq.REQ)
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(zurl)
        log.debug(socket)
        payload = {
            'message': 'ping',
            'timestamp': time.time()
        }
        current_timeout = int((timeout + (failures and timeout or 0) * (failures**backoff)))
        log.debug('Pinging %s , payload: %s, timeout: %d' % (zurl, payload, current_timeout))

        await socket.send(msgpack.packb(payload))

        if await socket.poll(current_timeout):  # Success condition
            reply = await socket.recv()  # Data is ready for snarfing

            log.debug("ping success: %s : %s" % (
                zurl, msgpack.unpackb(reply, encoding='utf-8')))
            socket.close()

            active_state[record['mercury_id']]['last_ping'] = time.time()
            active_state[record['mercury_id']]['pinging'] = False
            return

        log.debug('ping timeout failures/backoff: {}/{}  : {} : {}'.format(
            failures, backoff, record['mercury_id'], zurl))

        failures += 1

        socket.close()

    log.info('{} timed out, removing active record'.format(record['mercury_id']))

    # remove agent data from the database
    # this will await until the inventory is available
    await inventory_client.update_one(
        record['mercury_id'], {'active': None})

    # remove the record from the state data structure
    del active_state[record['mercury_id']]

    # mark any active associated tasks as failed
    # retrieve currently active tasks
    result = await rpc_client.get_active_tasks_by_mercury_id(record['mercury_id'])
    task_count = result['message']['count']
    # error tasks out
    if task_count > 0:
        tasks = result['message']['tasks']
        for task in tasks:
            await rpc_client.complete_task({
                'job_id': task['job_id'],
                'task_id': task['task_id'],
                'status': 'ERROR',
                'message': "AAWOL - Agent went AWOL while handling task.",
            })


async def ping_loop(ctx,
                    ping_interval,
                    cycle_time,
                    initial_ping_timeout,
                    ping_retries,
                    backoff,
                    loop,
                    inventory_router_url,
                    rpc_client):
    """

    :param ctx:
    :param ping_interval:
    :param cycle_time:
    :param initial_ping_timeout:
    :param ping_retries:
    :param backoff:
    :param loop:
    :param inventory_router_url:
    :return:
    """
    # load the queue
    inventory_client = InventoryClient(inventory_router_url, linger=0)
    inventory_client.service_name = 'Inventory Service'

    while True:
        if stop_ping_loop:
            log.info('Stopping ping loop')
            break
        now = time.time()
        for mercury_id, data in list(active_state.items()):  # copy to list because the list length could change
            # out from under us
            if now - data['last_ping'] > ping_interval and not data['pinging']:
                log.debug('Scheduling ping for {}'.format(mercury_id))
                active_state[mercury_id]['pinging'] = True
                asyncio.ensure_future(ping(data, ctx, initial_ping_timeout, ping_retries,
                                           backoff, inventory_client, rpc_client),
                                      loop=loop)
        await asyncio.sleep(cycle_time)



import json
import operator
from functools import reduce

from behave import given, when, step
from src.tests.behave.api.features.common.utils import (
    get_entity_list_container_field,
    get_entity_id_field,
    read_json_from_file,
)


@given(
    "I have query details in {filename} for entities using the {service_name} api"
)
def step_i_have_query_details_in_filename_for_entities_using_the_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type filename: str
    :type service_name: str
    """
    location = context.json_location
    data = read_json_from_file(filename, location)

    context.services[service_name]["details"]["query"] = data


@when("I get the query_results from a query of {service_name}")
def step_i_get_the_query_results_from_a_query_of_service_api(
    context, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["query"]

    context.services[service_name]["resp"] = service_client.post(
        data=json.dumps(data), url_suffix="query"
    )


@when(
    "I get with bad headers in {filename} the query_results from a query of {service_name}"
)
def step_i_get_the_query_results_from_a_query_of_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """
    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["query"]

    location = context.json_location
    headers = read_json_from_file(filename, location)

    context.services[service_name]["resp"] = service_client.post(
        data=json.dumps(data), url_suffix="query", headers=headers
    )


@when(
    "I get with parameters in {filename} the query_results from a query of {service_name}"
)
def step_i_get_the_query_results_from_a_query_with_params_in_filename_of_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type filename: str
    :type service_name: str
    """

    location = context.json_location
    param_data = read_json_from_file(filename, location)

    keys = param_data.keys()
    suffix = "query?"
    for key in keys:
        suffix = "{0}{1}={2}&".format(suffix, key, param_data[key])
    # trim trailing &
    suffix = suffix.rstrip("&")
    context.services[service_name]["param_data"] = param_data

    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["query"]

    context.services[service_name]["resp"] = service_client.post(
        data=json.dumps(data), url_suffix=suffix
    )


@step(
    "I get with offset parameters in {filename} the query_results from a query of {service_name}"
)
def step_i_get_the_offset_query_results_from_a_query_with_params_in_filename_of_service_api(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type filename: str
    :type service_name: str
    """

    service_resp = context.services[service_name]["resp"]
    container_field = get_entity_list_container_field(service_name)
    service_entities = service_resp.json()[container_field]
    # TODO configure or calculate the indices
    offset_id = service_entities[4]["_id"]
    first_id = service_entities[5]["_id"]
    context.services[service_name]["offset_id"] = offset_id
    context.services[service_name]["first_id"] = first_id

    location = context.json_location
    param_data = read_json_from_file(filename, location)

    keys = param_data.keys()
    suffix = "query?"
    for key in keys:
        if key == "offset_id":
            param_data["offset_id"] = offset_id
        suffix = "{0}{1}={2}&".format(suffix, key, param_data[key])
    # trim trailing &
    suffix = suffix.rstrip("&")
    context.services[service_name]["offset_param_data"] = param_data

    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["query"]

    context.services[service_name]["offset_resp"] = service_client.post(
        data=json.dumps(data), url_suffix=suffix
    )
    # so we can use existing checks on this response too
    context.services[service_name]["resp"] = context.services[service_name][
        "offset_resp"
    ]


@step(
    "the {service_name} entities in the response contain the data from {filename}"
)
def step_the_entities_in_the_response_contain_the_data_from_filename(
    context, service_name, filename
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type field: str
    :type value: str
    """
    service_client = context.services[service_name]["client"]
    service_resp = context.services[service_name]["resp"]
    container_field = get_entity_list_container_field(service_name)
    service_entities = service_resp.json()[container_field]
    # TODO if len is 0 add a computer somehow
    context.check.assertGreater(len(service_entities), 0)

    field_name = get_entity_id_field(service_name)
    query_data = context.services[service_name]["details"]["query"]
    for entity in service_entities:
        # Make a call for each returned entity
        entity_id = entity[field_name]
        resp = service_client.get(entity_id)

        query = query_data["query"]
        for key in query.keys():
            # find the expected value for each key in the query
            value = query[key]
            # find the actual value from the http response
            keys = key.split(".")
            try:
                actual_value = reduce(operator.getitem, keys, resp.json())
                context.check.assertEqual(value, actual_value)
            except TypeError:
                # The computer doesn't have an "active" dictionary to check
                # the values of, it likely went inactive after the initial query
                pass


# Used for bad method testing
@when("I query with {method} for {service_name}")
def step_i_use_method_on_query_service(context, method, service_name):
    """
    :type context: behave.runner.Context
    :type method: str
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    data = context.services[service_name]["details"]["query"]

    if method == "get":
        context.services[service_name]["resp"] = service_client.get(
            url_suffix="query"
        )
    elif method == "post":
        context.services[service_name]["resp"] = service_client.post(
            data=json.dumps(data), url_suffix="query"
        )
    # TODO etc



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Generic mongodb facilities"""

import bson
import logging
import pymongo


log = logging.getLogger(__name__)


def get_connection(server_or_servers=None, replica_set=None,
                   username=None, password=None):
    """Creates a new MongoDB client.

    :param server_or_servers: Hostname / MongoDB URI or list of hostnames /
        URIs for the client to connect to.
    :param replica_set: The name of the replica set to connect to.
    :param username:
    :param password:
    :returns: A MongoClient instance.
    """
    servers = server_or_servers
    if servers is not None:
        if not isinstance(servers, list):
            servers = [servers]

    log.info('Connecting to %s : replicaSet: %s' % (servers, replica_set))

    if username:
        return pymongo.MongoClient(servers, replicaset=replica_set,
                                   username=username, password=password)
    else:
        return pymongo.MongoClient(servers, replicaset=replica_set)


class MongoCollection(object):
    def __init__(self,
                 database,
                 collection,
                 server_or_servers=None,
                 replica_set=None):
        """Get a collection from a Mongo database.

        :param database: The name of the database containing the collection.
        :param collection: The name of the collection.
        :param server_or_servers: Hostname / MongoDB URI or list of hostnames /
            URIs for the client to connect to.
        :param replica_set: The name of the replica set to connect to.
        :raises: InvalidName if the database or collection name is invalid.
        """
        self.servers = server_or_servers
        self.replica_set = replica_set
        self.database_name = database
        self.collection_name = collection

        self.connection = get_connection(self.servers, self.replica_set)

        log.debug('database: %s' % self.database_name)
        self.db = self.connection[self.database_name]
        log.debug('collection: %s' % self.collection_name)
        self.collection = self.db[self.collection_name]


def get_collection(database,
                   collection,
                   connection=None,
                   server_or_servers=None,
                   replica_set=None):
    """Get a collection from a Mongo database.

    :param database: The name of the database containing the collection.
    :param collection: The name of the collection.
    :param connection: A MongoClient instance.
    :param server_or_servers: Hostname / MongoDB URI or list of hostnames /
        URIs for the client to connect to.
    :param replica_set: The name of the replica set to connect to.
    :returns: A MongoDB collection.
    :raises: InvalidName if the database or collection name is invalid.
    """
    database_name = database
    collection_name = collection

    if not connection:
        connection = get_connection(server_or_servers, replica_set)

    log.info('database: %s' % database_name)
    db = connection[database_name]
    log.info('collection: %s' % collection_name)
    return db[collection_name]


def serialize_object_id(obj):
    if isinstance(obj, bson.ObjectId):
        obj = str(obj)

    if isinstance(obj, dict):
        if '_id' in obj:
            obj['_id'] = str(obj['_id'])
    return obj


def deserialize_object_id(obj):
    if '_id' in obj:
        if isinstance(obj['_id'], dict):
            # ex: {'_id': {'$gt': 'xxxxx'}}
            for key, value in list(obj['_id'].items()):
                if bson.ObjectId.is_valid(value):
                    obj['_id'][key] = bson.ObjectId(value)
                    break  # there should only be one
        elif isinstance(obj['_id'], str):
            obj['_id'] = bson.ObjectId(obj['_id'])
    return obj
